# Scola version 1.1.0
# ScolaStepper.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of ScolaWizard.
# ScolaWizard is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# ScolaWizard is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with ScolaWizard.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Simulation Management
# 4) Process creation and deletion
# 5) Process actions
# 6) Instructions
# 7) Assertions
# 8) Backtrack
# 9) Print functions
# 10) Messages

# 1) Imported modules
# -------------------

import sys
import os

from ScolaModels import *
from ScolaHistories import *
from ScolaProcesses import *
from ScolaExpressionEvaluator import *
from ScolaBacktrackStack import *
from ScolaParser import *
from ScolaPrinter import *
from ScolaRewriter import *

# 2) Creator
# ----------

class ScolaStepper:
  def __init__(self):
    self._evaluator = ScolaExpressionEvaluator()
    self._backtrackStack = ScolaBacktrackStack()
    self._backtrackLimit = 0
    self._processes = []
    self._model = ScolaModel()
    self._compiledModel = None
    self._outputFile = sys.stdout
    self._errorFile = sys.stderr
    self._parser = ScolaParser()
    self._rewriter = ScolaRewriter()
    self._printer = ScolaPrinter()
    self._printer.SetOutputFile(self._outputFile)
    self._simulationStarted = False
    self._verbose = True
    self._modification = True
    self._fixpointReached = True

# 3) Simulation management
# ------------------------

  def ResetModel(self):
    if self._simulationStarted:
      self.StopSimulation()
    self._model = ScolaModel()
    ScolaUniverse.Reset()
    self._compiledModel = None
    self._backtrackLimit = 0

  def ParseFile(self, inputFileName, inputFile):
    self._compiledModel = None
    self._parser.SetModel(self._model)
    numberOfErrors = self._parser.ParseFile(inputFileName, inputFile)
    return numberOfErrors

  def IsModelCompiled(self):
    return self._compiledModel!=None
    
  def CompileModel(self):
    if self._compiledModel!=None:
      return 0
    self._compiledModel = self._rewriter.CompileModel(self._model)
    if self._rewriter.GetNumberOfErrors()>0:
      self._compiledModel = None
    return self._rewriter.GetNumberOfErrors()

  def IsSimulationStarted(self):
    return self._simulationStarted

  def StartSimulation(self):
    if self._simulationStarted:
      self.RaiseError("Simulation already started")
      return False
    self._simulationStarted = True
    self._evaluator.SetModel(self._compiledModel)
    self.ResetPortsOfModelOnInitialValues()
    self.ApplyAssertions()
    self._processes = []
    self._backtrackStack.Reset()
    self.CollectInitialStates()
    self._backtrackLimit = 0
    return True

  def StopSimulation(self):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return False
    self._simulationStarted = False
    self._processes = []
    self._backtrackStack.Reset()
    return True

  def ResetPortsOfModelOnInitialValues(self):
    if self._compiledModel==None:
      return
    for element in self._compiledModel.GetElements():
      self.ResetPortsOfElementOnInitialValues(element)

  def ResetPortsOfElementOnInitialValues(self, element):
    if element.GetType()==ScolaModel.BLOCK:
      for child in element.GetElements():
        self.ResetPortsOfElementOnInitialValues(child)
    elif element.GetType()==ScolaModel.PORT:
      systemPath = self._evaluator.GetSystemPath()
      self._evaluator.SetSystemPath(element.GetSystemPath())
      initialization = element.GetInitialization()
      value = self._evaluator.EvalExpression(initialization)
      element.SetValue(value)
      self._evaluator.SetSystemPath(systemPath)

# 4) Process creation and deletion
# --------------------------------

  def GetNumberOfProcesses(self):
    return len(self._processes)

  def GetProcess(self, number):
    if number<1 or number>self.GetNumberOfProcesses():
      return None
    return self._processes[number-1]

  def GetProcessFromNumber(self, number):
    process = self.GetProcess(number)
    if process==None:
      self.RaiseError("%d is not a process number" % number)
      return None
    return process

  def GetNumbersOfActiveProcesses(self):
    return [process.GetNumber() for process in self._processes if process.IsActive()]

  def NewProcess(self, locationName):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return -1
    element = self.GetScenarioElementFromString(locationName)
    if element==None: return -1
    if element.GetType()!=ScolaModel.STATE:
      self.RaiseError("Processes must created on states")
      return -1
    command = ScolaCommand.NewNewProcess(element.GetAbsolutePath().ToString())
    self._backtrackStack.PushBlock(command)
    process = self._NewProcess(element)
    command.SetComment("create-process " + str(process.GetNumber()))
    return process.GetNumber()

  def _NewProcess(self, element):
    process = ScolaProcess(self.GetNumberOfProcesses()+1, element)
    self._processes.append(process)
    self._backtrackStack.NewNewProcess(process)
    self._ActivateProcess(process)
    return process

  def _ActivateProcess(self, process):
    if process.IsActive():
      return
    process.SetActive(True)
    self._backtrackStack.NewActivateProcess(process)
    location = process.GetLocation()
    if location.GetType()==ScolaModel.BRANCH:
      location.EnqueueProcess(process)
      self._backtrackStack.NewEnqueueProcess(process, location)

  def DeactivateProcesses(self, processNumbers):
    count = 0
    for processNumber in processNumbers:
      ok = self.DeactivateProcess(processNumber)
      if ok:
        count += 1
    return count

  def DeactivateProcess(self, processNumber):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return False
    process = self.GetProcessFromNumber(processNumber)
    if process==None:
      return False
    self._backtrackStack.PushBlock(ScolaCommand.NewKillProcess(process.GetNumber()))
    self._DeactivateProcess(process)
    return True

  def _DeactivateProcess(self, process):
    if not process.IsActive():
      return
    process.SetActive(False)
    element = process.GetLocation()
    if element.GetType()==ScolaModel.BRANCH:
      index = element.RemoveProcess(process)
      self._backtrackStack.NewDequeueProcess(process, element, index)
    self._backtrackStack.NewDeactivateProcess(process, process.GetLocation())

  def GetScenarioElementFromString(self, string):
    if not ScolaPath.IsValidPath(string):
      self.RaiseError("%s is not a valid path" % string)
      return None
    path = ScolaPath.FromString(string)
    element = self._compiledModel.AccessNamedElement(path)
    if element==None or not ScolaModel.IsScenarioElement(element):
      self.RaiseError("%s is not a scenario element" % path.ToString())
      return None
    return element

  def CollectInitialStates(self):
    states = []
    self._compiledModel.CollectElements(lambda element: element.GetType()==ScolaModel.STATE and element.IsInitial(), states)
    return states

  def GetPathsOfInitialStates(self):
    return [state.GetAbsolutePath().ToString() for state in self.CollectInitialStates()]

  def IsProcessLocatedOnATerminalState(self, processNumber):
    process = self.GetProcessFromNumber(processNumber)
    if process==None:
      return False
    location = process.GetLocation()
    return location.GetType()==ScolaModel.STATE and location.GetNextElement()==None

  def IsProcessLocatedOnAChoiceGateway(self, processNumber):
    process = self.GetProcessFromNumber(processNumber)
    if process==None:
      return False
    location = process.GetLocation()
    return location.GetType()==ScolaModel.CHOICE

  def GetBranchesOfLocationOfProcess(self, processNumber):
    process = self.GetProcessFromNumber(processNumber)
    if process==None:
      return []
    location = process.GetLocation()
    if location.GetType()!=ScolaModel.CHOICE:
      return []
    # return [branch.GetAbsolutePath().ToString() for branch in location.GetElements()]
    return [branch.GetIdentifier() for branch in location.GetElements()]

# 5) Process actions
# ------------------

  def MoveProcessForward(self, processNumber, branchName):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return False
    process = self.GetProcessFromNumber(processNumber)
    if process==None:
      return False
    location = process.GetLocation()
    if location.GetType()!=ScolaModel.CHOICE or branchName==None:
      childElement = None
    else:
      childElement = location.LookForNamedElement(branchName)
    return self._MoveProcessForward(process, childElement)

  def _MoveProcessForward(self, process, childElement):
    if not process.IsActive():
      self.RaiseError("Process %d is not active" % process.GetNumber())
      return False
    sourceElement = process.GetLocation()

    if sourceElement.GetType()==ScolaModel.STATE:
      nextElement = self.GetProcessNextElement(process)
      if nextElement==None:
        return False
      command = ScolaCommand.NewNext(process.GetNumber())
      command.SetComment("goto " + self._printer.GetStringOfScenarioElement(nextElement))
      self._backtrackStack.PushBlock(command)
      self._MoveProcess(process, nextElement)

    elif sourceElement.GetType()==ScolaModel.TASK:
      nextElement = self.GetProcessNextElement(process)
      if nextElement==None:
        return False
      command = ScolaCommand.NewNext(process.GetNumber())
      self._backtrackStack.PushBlock(command)
      ok = True
      for instruction in sourceElement.GetElements():
        ok = self.ExecuteInstruction(instruction)
        if not ok:
          break
      if ok:
        ok = self.ApplyAssertions()
      if not ok:
        self.RaiseError("Unable to complete task %s" % sourceElement.GetAbsolutePath().ToString())
        self.Backtrack()
        return False
      command.SetComment("goto " + self._printer.GetStringOfScenarioElement(nextElement))
      self._MoveProcess(process, nextElement)

    elif sourceElement.GetType()==ScolaModel.TEST:
      found = False
      nextElement = None
      for caseElement in sourceElement.GetElements():
        condition = caseElement.GetCondition()
        self._evaluator.SetSystemPath(sourceElement.GetSystemPath())
        value = self._evaluator.EvalExpression(condition)
        self._evaluator.SetSystemPath(None)
        if value.IsTrue():
          found = True
          nextElement = caseElement.GetNextElement()
          break
        if not value.IsFalse():
          self.RaiseError("Condition of case %s cannot be evaluated" % caseElement.GetAbsolutePath().ToString())
          return False
      if not found:
        self.RaiseError("None of the case of gateway %s is evaluated to true" % sourceElement.GetAbsolutePath().ToString())
        return False
      if nextElement==None:
        self.RaiseError("Case %s is terminal" % caseElement.GetAbsolutePath().ToString())
        return False
      command = ScolaCommand.NewNext(process.GetNumber())
      command.SetComment("goto " + self._printer.GetStringOfScenarioElement(nextElement) + " via " + self._printer.GetStringOfScenarioElement(caseElement))
      self._backtrackStack.PushBlock(command)
      self._MoveProcess(process, nextElement)

    elif sourceElement.GetType()==ScolaModel.CHOICE:
      if childElement==None:
        self.RaiseError("Next branch of gateway %s must be specified" % sourceElement.GetAbsolutePath().ToString())
        return False
      if childElement.GetContainer()!=sourceElement:
        self.RaiseError("%s is not a branch of gateway %s" % childElement.GetAbsolutePath().ToString(), sourceElement.GetAbsolutePath().ToString())
        return False
      nextElement = childElement.GetNextElement()
      if nextElement==None:
        path = childElement.GetAbsolutePath()
        self.RaiseError("%s is terminal" % path.ToString())
        return False
      command = ScolaCommand.NewChoice(process.GetNumber(), childElement.GetIdentifier())
      command.SetComment("goto " + self._printer.GetStringOfScenarioElement(nextElement))
      self._backtrackStack.PushBlock(command)
      self._MoveProcess(process, nextElement)

    elif sourceElement.GetType()==ScolaModel.FORK:
      command = ScolaCommand.NewNext(process.GetNumber())
      self._backtrackStack.PushBlock(command)
      comment = "deactivate-process " + str(process.GetNumber())
      for branch in sourceElement.GetElements():
        nextElement = branch.GetNextElement()
        if nextElement==None:
          self.Backtrack()
          self.RaiseError("%s is terminal" % branch.GetAbsolutePath().ToString())
          return False
        childProcess = self._NewProcess(nextElement)
        # process.AddChild(childProcess)
        comment += "\n\t#create-process " + str(childProcess.GetNumber()) + " at " + self._printer.GetStringOfScenarioElement(nextElement) 
      self._DeactivateProcess(process)
      command.SetComment(comment)

    elif sourceElement.GetType()==ScolaModel.SPLIT:
      command = ScolaCommand.NewNext(process.GetNumber())
      self._backtrackStack.PushBlock(command)
      comment = "deactivate-process " + str(process.GetNumber())
      for branch in sourceElement.GetElements():
        nextElement = branch.GetNextElement()
        if nextElement==None:
          self.Backtrack()
          self.RaiseError("%s is terminal" % branch.GetAbsolutePath().ToString())
          return False
        childProcess = self._NewProcess(nextElement)
        process.AddChild(childProcess)
        comment += "\n\t#create-process " + str(childProcess.GetNumber()) + " at " + self._printer.GetStringOfScenarioElement(nextElement) 
      self._DeactivateProcess(process)
      command.SetComment(comment)

    elif sourceElement.GetType()==ScolaModel.BRANCH:
      container = sourceElement.GetContainer()

      if container.GetType()==ScolaModel.JOIN:
        nextElement = self.GetProcessNextElement(process)
        if nextElement==None:
          return False
        index = sourceElement.GetIndexInQueue(process)
        if index!=0:
          self.RaiseError("%d is not the first process in the queue " % process.GetNumber())
          return False
        enabled = True
        for branch in container.GetElements():
          if branch.IsQueueEmpty():
            self.RaiseError("Waiting for a process on " + branch.GetAbsolutePath().ToString())
            enabled = False
        if not enabled:
          return False
        command = ScolaCommand.NewNext(process.GetNumber())
        self._backtrackStack.PushBlock(command)
        comment = ""
        first = True
        for branch in container.GetElements():
          joinedProcess = branch.FirstInQueue()
          self._DeactivateProcess(joinedProcess)
          if first:
            first = False
          else:
            comment += "\t#"
          comment += "kill-process " + str(joinedProcess.GetNumber()) + "\n"
        newProcess = self._NewProcess(nextElement)
        comment += "\t#create-process " + str(newProcess.GetNumber()) + " at " + self._printer.GetStringOfScenarioElement(nextElement)
        command.SetComment(comment)

      elif container.GetType()==ScolaModel.MERGE:
        nextElement = self.GetProcessNextElement(process)
        if nextElement==None:
          return False
        parentProcess = process.GetParent()
        if parentProcess==None:
          self.RaiseError("The process %d has no parent process" % process.GetNumber())
          return False
        ok = True
        for childProcess in parentProcess.GetChildren():
          location = childProcess.GetLocation()
          locationContainer = location.GetContainer()
          if locationContainer!=container:
            self.RaiseError("Waiting for process %d" % childProcess.GetNumber())
            print("[%d] Process %d is on %s" % (process.GetNumber(), childProcess.GetNumber(), self._printer.GetStringOfScenarioElement(location)))
            ok = False
        if not ok:
          return False
        command = ScolaCommand.NewNext(process.GetNumber())
        self._backtrackStack.PushBlock(command)
        comment = ""
        first = True
        for childProcess in parentProcess.GetChildren():
          self._DeactivateProcess(childProcess)
          if first:
            first = False
          else:
            comment += "\t#"
          comment += "kill-process " + str(childProcess.GetNumber()) + "\n"
        parentProcess.SetLocation(nextElement)
        self._ActivateProcess(parentProcess)
        comment += "\t#reactivate-process " + str(parentProcess.GetNumber()) + " at " + self._printer.GetStringOfScenarioElement(nextElement)
        command.SetComment(comment)

      elif container.GetType()==ScolaModel.MEET:
        index = sourceElement.GetIndexInQueue(process)
        if index!=0:
          self.RaiseError("%d is not the first process in the queue " % process.GetNumber())
          return False
        enabled = True
        for branch in container.GetElements():
          if branch.IsQueueEmpty():
            self.RaiseError("Waiting for a process on " + branch.GetAbsolutePath().ToString())
            enabled = False
            continue
          nextElement = branch.GetNextElement()
          if nextElement==None:
            path = nextElement.GetAbsolutePath()
            self.RaiseError("%s is terminal" % path.ToString())
            enabled = False
            continue
        if not enabled:
          return False
        for branch in container.GetElements():
          attendee = branch.DequeueProcess()
          self._backtrackStack.NewDequeueProcess(attendee, branch, 0)
          nextElement = branch.GetNextElement()
          self._MoveProcess(attendee, nextElement)

    else:
      path = sourceElement.GetAbsolutePath()
      self.RaiseError("Gateway %s not implemented yet" % path.ToString())
      return False
    return True

  def _MoveProcess(self, process, nextElement):
    self._backtrackStack.NewMoveProcess(process, process.GetLocation())
    process.SetLocation(nextElement)
    if nextElement.GetType()==ScolaModel.BRANCH:
      self._backtrackStack.NewEnqueueProcess(process, nextElement)
      nextElement.EnqueueProcess(process)

  def GetProcessNextElement(self, process):
    element = process.GetLocation()
    if element.GetType()==ScolaModel.BRANCH:
      container = element.GetContainer()
      nextElement = container.GetNextElement()
    else:
      nextElement = element.GetNextElement()
    if nextElement==None:
      path = element.GetAbsolutePath()
      self.RaiseError("%s is terminal" % path.ToString())
      return None
    return nextElement

# 6) Instructions
# ---------------

  def ExecuteInstruction(self, instruction):
    if instruction.GetType()==ScolaModel.FAIL:
      ok = self.ExecuteFailInstruction(instruction)
    elif instruction.GetType()==ScolaModel.SET:
      ok = self.ExecuteSetInstruction(instruction)
    elif instruction.GetType()==ScolaModel.NEW_PORT:
      ok = self.ExecuteNewPortInstruction(instruction)
    elif instruction.GetType()==ScolaModel.NEW_BLOCK:
      ok = self.ExecuteNewBlockInstruction(instruction)
    elif instruction.GetType()==ScolaModel.DELETE:
      ok = self.ExecuteDeleteInstruction(instruction)
    elif instruction.GetType()==ScolaModel.MOVE:
      ok = self.ExecuteMoveInstruction(instruction)
    elif instruction.GetType()==ScolaModel.SEND:
      ok = self.ExecuteSendInstruction(instruction)
    elif instruction.GetType()==ScolaModel.RECEIVE:
      ok = self.ExecuteReceiveInstruction(instruction)
    elif instruction.GetType()==ScolaModel.CLONE:
      ok = self.ExecuteCloneInstruction(instruction)
    elif instruction.GetType()==ScolaModel.IF:
      ok = self.ExecuteIfInstruction(instruction)
    elif instruction.GetType()==ScolaModel.BEGIN:
      ok = self.ExecuteInstructionBlock(instruction)
    elif instruction.GetType()==ScolaModel.PRINT:
      ok = self.ExecutePrintInstruction(instruction)
    else:
      ok = False
    return ok

  def ExecuteFailInstruction(self, instruction):
    return False

  def ExecuteSetInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    path = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if path==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    element = self._compiledModel.AccessNamedElement(path)
    if element==None or element.GetType()!=ScolaModel.PORT:
      self.RaiseError("%s is not a port" % path.ToString())
      return False
    port = element
    self._evaluator.SetSystemPath(systemPath)
    newValue = self._evaluator.EvalExpression(instruction.GetValueExpression())
    self._evaluator.SetSystemPath(None)
    baseType = port.GetBaseType()
    if not baseType.Contains(newValue):
      self.RaiseError("Attempt to set non authorized value (%s) to port %s" % (ScolaConstant.ToString(newValue).GetValueAsString(), path.ToString()))
      return False
    oldValue = port.GetValue()
    if ScolaConstant.AreDf(newValue, oldValue):
      self._backtrackStack.NewSetPort(port, oldValue)
      port.SetValue(newValue)
      self._modification = True
    return True

  def ExecuteNewPortInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    blockPath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if blockPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    identifier = blockPath.RemoveLastIdentifier()
    element = self._compiledModel.AccessNamedElement(blockPath)
    if element==None or (element.GetType()!=ScolaModel.MODEL and element.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % blockPath.ToString())
      return False
    parentBlock = element
    port = parentBlock.LookForNamedElement(identifier)
    if port!=None:
      self.RaiseError("%s already exists" % port.GetAbsolutePath().ToString())
      return False
    expression = instruction.GetInitializationExpression()
    self._evaluator.SetSystemPath(systemPath)
    value = self._evaluator.EvalExpression(expression)
    self._evaluator.SetSystemPath(None)
    port = ScolaPort(instruction.GetBaseType(), identifier, expression, parentBlock)
    port.SetValue(value)
    port.SetSystemPath(blockPath)
    self._backtrackStack.NewNewElement(port)
    return True

  def ExecuteNewBlockInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    blockPath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if blockPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    identifier = blockPath.RemoveLastIdentifier()
    element = self._compiledModel.AccessNamedElement(blockPath)
    if element==None or (element.GetType()!=ScolaModel.MODEL and element.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % blockPath.ToString())
      return False
    parentBlock = element
    block = parentBlock.LookForNamedElement(identifier)
    if block!=None:
      self.RaiseError("%s already exists" % block.GetAbsolutePath().ToString())
      return False
    blockClass = instruction.GetBlockClass()
    if blockClass==None:
      block = ScolaBlock(identifier, parentBlock)
      self._backtrackStack.NewNewElement(block)
      return True
    self.RaiseError("Dynamic create of instances of classes is not implemented yet")
    return False

  def ExecuteDeleteInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    blockPath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if blockPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    identifier = blockPath.RemoveLastIdentifier()
    element = self._compiledModel.AccessNamedElement(blockPath)
    if element==None or (element.GetType()!=ScolaModel.MODEL and element.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % blockPath.ToString())
      return False
    parentBlock = element
    element = parentBlock.LookForNamedElement(identifier)
    if element==None:
      self.RaiseError("%s.%s does not exist" % (blockPath.ToString(), identifier))
      return False
    index = parentBlock.RemoveElement(element)
    self._backtrackStack.NewDeleteElement(parentBlock, element, index)
    return True

  def ExecuteMoveInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    sourcePath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if sourcePath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    sourceIdentifier = sourcePath.RemoveLastIdentifier()
    element = self._compiledModel.AccessNamedElement(sourcePath)
    if element==None or (element.GetType()!=ScolaModel.MODEL and element.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % sourcePath.ToString())
      return False
    sourceBlock = element
    sourceElement = sourceBlock.LookForNamedElement(sourceIdentifier)
    if sourceElement==None:
      self.RaiseError("%s.%s does not exist" % (sourcePath.ToString(), sourceIdentifier))
      return False
    self._evaluator.SetSystemPath(systemPath)
    targetPath = self._evaluator.EvalExpressionAsPath(instruction.GetTargetPathExpression())
    self._evaluator.SetSystemPath(None)
    if targetPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetTargetPathExpression()))
      return False
    targetIdentifier = targetPath.RemoveLastIdentifier()
    targetElement = self._compiledModel.AccessNamedElement(targetPath)
    if targetElement==None or (targetElement.GetType()!=ScolaModel.MODEL and targetElement.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % targetPath.ToString())
      return False
    targetBlock = targetElement
    targetElement = targetBlock.LookForNamedElement(targetIdentifier)
    if targetElement!=None:
      self.RaiseError("%s[%s].%s already exists" % (targetPath.ToString(), targetBlock.GetAbsolutePath().ToString(), targetIdentifier))
      return False
    index = sourceBlock.RemoveElement(sourceElement)
    sourceElement.SetIdentifier(targetIdentifier)
    targetBlock.AddElement(sourceElement)
    self._backtrackStack.NewMoveElement(sourceBlock, sourceElement, index, sourceIdentifier)
    return True

  def ExecuteSendInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    sourcePath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if sourcePath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    sourceElement = self._compiledModel.AccessNamedElement(sourcePath)
    if sourceElement==None or (sourceElement.GetType()!=ScolaModel.BLOCK and sourceElement.GetType()!=ScolaModel.PORT):
      self.RaiseError("%s is not a block or a port" % sourcePath.ToString())
      return False
    self._evaluator.SetSystemPath(systemPath)
    receiverPath = self._evaluator.EvalExpressionAsPath(instruction.GetReceiverPathExpression())
    self._evaluator.SetSystemPath(None)
    if receiverPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetReceiverPathExpression()))
      return False
    receiver = self._compiledModel.AccessNamedElement(receiverPath)
    if receiver==None or receiver.GetType()!=ScolaModel.BLOCK:
      self.RaiseError("%s is not a block" % receiverPath.ToString())
      return False
    sourceBlock = sourceElement.GetContainer()
    self._evaluator.SetSystemPath(ScolaPath.NewEmptyPath())
    keyPath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    key = sourceBlock.GetAbsolutePath().ToString() + ":" + keyPath.ToString()
    targetElement = receiver.LookForElementInReceptionSet(key)
    if targetElement!=None:
      self.RaiseError("%s has already an element %s waiting for reception" % (receiver.GetAbsolutePath().ToString(), key))
      return False
    index = sourceBlock.RemoveElement(sourceElement)
    receiver.AddElementToReceptionSet(key, sourceElement)
    self._backtrackStack.NewSendElement(sourceBlock, sourceElement, index, receiver, key)
    return True

  def ExecuteReceiveInstruction(self, instruction):
    receiverPath = instruction.GetSystemPath()
    receiver = self._compiledModel.AccessNamedElement(receiverPath)
    if receiver==None or receiver.GetType()!=ScolaModel.BLOCK:
      self.RaiseError("%s is not a block" % receiverPath.ToString())
      return False
    self._evaluator.SetSystemPath(ScolaPath.NewEmptyPath())
    keyPath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if keyPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    self._evaluator.SetSystemPath(receiverPath)
    senderPath = self._evaluator.EvalExpressionAsPath(instruction.GetSenderPathExpression())
    self._evaluator.SetSystemPath(None)
    if senderPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetSenderPathExpression()))
      return False
    sender = self._compiledModel.AccessNamedElement(senderPath)
    if sender==None or sender.GetType()!=ScolaModel.BLOCK:
      self.RaiseError("%s is not a block" % senderPath.ToString())
      return False
    key = sender.GetAbsolutePath().ToString() + ":" + keyPath.ToString()
    sourceElement = receiver.LookForElementInReceptionSet(key)
    if sourceElement==None:
      self.RaiseError("%s has no element %s waiting for reception" % (receiverPath.ToString(), key))
      return False
    self._evaluator.SetSystemPath(receiverPath)
    targetPath = self._evaluator.EvalExpressionAsPath(instruction.GetTargetPathExpression())
    self._evaluator.SetSystemPath(None)
    if targetPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetTargetPathExpression()))
      return False
    targetIdentifier = targetPath.RemoveLastIdentifier()
    targetElement = self._compiledModel.AccessNamedElement(targetPath)
    if targetElement==None or targetElement.GetType()!=ScolaModel.BLOCK:
      self.RaiseError("%s is not a block" % targetPath.ToString())
      return False
    targetBlock = targetElement
    receiver.RemoveElementFromReceptionSet(key)
    sourceIdentifier = sourceElement.GetIdentifier()
    sourceElement.SetIdentifier(targetIdentifier)
    targetBlock.AddElement(sourceElement)
    self._backtrackStack.NewReceiveElement(sourceElement, sourceIdentifier, receiver, key)
    return True

  def ExecuteCloneInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    sourcePath = self._evaluator.EvalExpressionAsPath(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    if sourcePath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetPathExpression()))
      return False
    sourceIdentifier = sourcePath.RemoveLastIdentifier()
    element = self._compiledModel.AccessNamedElement(sourcePath)
    if element==None or (element.GetType()!=ScolaModel.MODEL and element.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % sourcePath().ToString())
      return False
    sourceBlock = element
    sourceElement = sourceBlock.LookForNamedElement(sourceIdentifier)
    if sourceElement==None:
      self.RaiseError("%s.%s does not exist" % (sourcePath.ToString(), sourceIdentifier))
      return False
    self._evaluator.SetSystemPath(systemPath)
    targetPath = self._evaluator.EvalExpressionAsPath(instruction.GetTargetPathExpression())
    self._evaluator.SetSystemPath(None)
    if targetPath==None:
      self.RaiseError("%s does not evaluate as a valid path" % self._printer.GetStringOfExpression(instruction.GetTargetPathExpression()))
      return False
    targetIdentifier = targetPath.RemoveLastIdentifier()
    targetElement = self._compiledModel.AccessNamedElement(targetPath)
    if targetElement==None or (targetElement.GetType()!=ScolaModel.MODEL and targetElement.GetType()!=ScolaModel.BLOCK):
      self.RaiseError("%s is not a block" % targetPath.ToString())
      return False
    targetBlock = targetElement
    targetElement = targetBlock.LookForNamedElement(targetIdentifier)
    if targetElement!=None:
      self.RaiseError("%s.%s already exists" % (targetPath.ToString(), targetIdentifier))
      return False
    if sourceElement.GetType()==ScolaModel.BLOCK:
      targetElement = targetBlock.CloneBlock(sourceElement, targetIdentifier)
    elif sourceElement.GetType()==ScolaModel.PORT:
      targetElement = targetBlock.ClonePort(sourceElement, targetIdentifier)
    self._backtrackStack.NewCloneElement(targetElement)
    return True

  def ExecuteIfInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    conditionValue = self._evaluator.EvalExpression(instruction.GetCondition())
    self._evaluator.SetSystemPath(None)
    if conditionValue.IsTrue():
      return self.ExecuteInstruction(instruction.GetThenInstruction())
    if conditionValue.IsFalse():
      if instruction.GetElseInstruction()!=None:
        return self.ExecuteInstruction(instruction.GetElseInstruction())
      return True
    self.RaiseError("%s does not evaluate" % self._printer.GetStringOfExpression(instruction.GetCondition()))
    return False

  def ExecuteInstructionBlock(self, instructionBlock):
    for instruction in instructionBlock.GetElements():
      ok = self.ExecuteInstruction(instruction)
      if not ok:
        return False
    return True

  def ExecutePrintInstruction(self, instruction):
    systemPath = instruction.GetSystemPath()
    self._evaluator.SetSystemPath(systemPath)
    value = self._evaluator.EvalExpression(instruction.GetPathExpression())
    self._evaluator.SetSystemPath(None)
    self._printer.PrintConstant(value)
    return True

# 7) Assertions
# -------------

  def ApplyAssertions(self):
    while True:
      self._fixpointReached = True
      ok = self._ApplyAssertions(self._compiledModel)
      if not ok:
        return False
      if self._fixpointReached:
        break
    return True

  def _ApplyAssertions(self, container):
    for element in container.GetElements():
      if element.GetType()==ScolaModel.BLOCK:
        ok = self._ApplyAssertions(element)
        if not ok:
          return False
      elif element.GetType()==ScolaModel.ASSERTION:
        ok = self._ApplyAssertion(element)
        if not ok:
          return False
    return True

  def _ApplyAssertion(self, assertion):
    for instruction in assertion.GetElements():
      self._modification = False
      if instruction.GetType()==ScolaModel.SET:
        ok = self.ExecuteSetInstruction(instruction)
        if not ok:
          return False
        if self._modification:
          self._fixpointReached = False
    return True

# 8) Backtrack
# ------------

  def GetBacktrackLimit(self):
    return self._backtrackLimit
  def SetBacktrackLimit(self, limit):
    self._backtrackLimit = limit
  def SetBacktrackLimitHere(self):
    self._backtrackLimit = self._backtrackStack.GetSize()

  def Backtrack(self):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return False
    if self._backtrackStack.IsEmpty() or self._backtrackStack.GetSize()<=self._backtrackLimit:
      self.RaiseError("No more action to backtrack")
      return False
    block = self._backtrackStack.PopBlock()
    while not block.IsEmpty():
      instruction = block.PopInstruction()
      self.BacktrackInstruction(instruction)
    return True

  def BacktrackAll(self):
    if not self._simulationStarted:
      self.RaiseError("Simulation is not started")
      return False
    while not self._backtrackStack.IsEmpty():
      self.Backtrack()
    return True

  def BacktrackInstruction(self, instruction):
    if instruction.GetType()==ScolaBacktrackInstruction.NEW_PROCESS:
      # print("backtrack new (%d)" % instruction.GetDatum(0).GetNumber())
      process = self._processes.pop()
      parentProcess = process.GetParent()
      if parentProcess!=None:
        parentProcess.RemoveChild(process)
      # print("%d processes remaining" % len(self._processes))
    elif instruction.GetType()==ScolaBacktrackInstruction.ACTIVATE_PROCESS:
      # print("backtrack activate %d" % instruction.GetProcess().GetNumber())
      process = instruction.GetDatum(0)
      process.SetActive(False)
    elif instruction.GetType()==ScolaBacktrackInstruction.DEACTIVATE_PROCESS:
      # print("backtrack deactivate %d" % instruction.GetProcess().GetNumber())
      process = instruction.GetDatum(0)
      location = instruction.GetDatum(1)
      process.SetLocation(location)
      process.SetActive(True)
    elif instruction.GetType()==ScolaBacktrackInstruction.MOVE_PROCESS:
      # print("backtrack move %d" % instruction.GetProcess().GetNumber())
      process = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      process.SetLocation(element)
    elif instruction.GetType()==ScolaBacktrackInstruction.ENQUEUE_PROCESS:
      process = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      element.RemoveProcess(process)
    elif instruction.GetType()==ScolaBacktrackInstruction.DEQUEUE_PROCESS:
      process = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      index = instruction.GetDatum(2)
      element.InsertProcess(index, process)
    elif instruction.GetType()==ScolaBacktrackInstruction.SET_PORT:
      # print("backtrack set port %s" % instruction.GetElement().GetIdentifier())
      port = instruction.GetDatum(0)
      value = instruction.GetDatum(1)
      port.SetValue(value)
    elif instruction.GetType()==ScolaBacktrackInstruction.NEW_ELEMENT:
      # print("backtrack new_... %s" % instruction.GetElement().GetIdentifier())
      element = instruction.GetDatum(0)
      container = element.GetContainer()
      container.RemoveElement(element)
    elif instruction.GetType()==ScolaBacktrackInstruction.DELETE_ELEMENT:
      # print("backtrack delete %s" % instruction.GetElement().GetIdentifier())
      container = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      index = instruction.GetDatum(2)
      container.InsertElement(index, element)
    elif instruction.GetType()==ScolaBacktrackInstruction.MOVE_ELEMENT:
      # print("backtrack move %s" % instruction.GetElement().GetIdentifier())
      sourceBlock = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      identifier = instruction.GetDatum(2)
      index = instruction.GetDatum(3)
      targetBlock = element.GetContainer()
      targetBlock.RemoveElement(element)
      element.SetIdentifier(identifier)
      sourceBlock.InsertElement(index, element)
      # print("backtrack move %s" % instruction.GetElement().GetIdentifier())
    elif instruction.GetType()==ScolaBacktrackInstruction.SEND_ELEMENT:
      sourceBlock = instruction.GetDatum(0)
      element = instruction.GetDatum(1)
      index = instruction.GetDatum(2)
      receiver = instruction.GetDatum(3)
      key = instruction.GetDatum(4)
      receiver.RemoveElementFromReceptionSet(key)
      sourceBlock.InsertElement(index, element)
    elif instruction.GetType()==ScolaBacktrackInstruction.RECEIVE_ELEMENT:
      # print("backtrack receive %s" % instruction.GetElement().GetIdentifier())
      element = instruction.GetDatum(0)
      identifier = instruction.GetDatum(1)
      receiver = instruction.GetDatum(2)
      key = instruction.GetDatum(3)
      container = element.GetContainer()
      container.RemoveElement(element)
      element.SetIdentifier(identifier)
      receiver.AddElementToReceptionSet(key, element)
    elif instruction.GetType()==ScolaBacktrackInstruction.CLONE_ELEMENT:
      # print("backtrack clone %s" % instruction.GetElement().GetIdentifier())
      element = instruction.instruction.GetDatum(0)
      container = element.GetContainer()
      container.RemoveElement(element)

  def GetHistory(self):
    history = ScolaHistory()
    for  index in range(0, self._backtrackStack.GetSize()):
      backtrackBlock = self._backtrackStack.GetBlock(index)
      command = backtrackBlock.GetCommand()
      history.AppendCommand(command)
    return history


# 9) Print functions
# ------------------

  def GetOutputFile(self):
    return self._outputFile
  def SetOutputFile(self, outputFile):
    self._outputFile = outputFile
    self._printer.SetOutputFile(self._outputFile)

  def GetErrorFile(self):
    return self._errorFile
  def SetErrorFile(self, _errorFile):
    self._errorFile = errorFile

  def GetIndentation(self):
    return self._printer.GetIndentation()
  def SetIndentation(self, indentation):
    self._printer.SetIndentation(indentation)

  def GetIdentifiersOfTopLevelBlocks(self):
    if self._compiledModel==None:
      return []
    identifiers = []
    for element in self._compiledModel.GetElements():
      if element.GetType()==ScolaModel.BLOCK:
        identifiers.append(element.GetIdentifier())
    return identifiers

  def PrintModel(self):
    self._printer.PrintModel(self._model)

  def PrintCompiledModel(self):
    if self._compiledModel==None:
      self.RaiseError("The model is not compiled yet")
      return
    self._printer.PrintModel(self._compiledModel)

  def PrintBlocks(self, pathStrings):
    for pathString in pathStrings:
      self.PrintBlock(pathString)

  def PrintBlock(self, pathString):
    if not ScolaPath.IsValidPath(pathString):
      self.RaiseError("%s is not a valid path" % pathString)
      return
    element = self.GetScenarioElementFromString(pathString)
    if element==None: return
    self._depth = 0
    self._printer.PrintValuesOfPortsOfBlock(element)

  def PrintProcesses(self, numbers):
    for number in numbers:
      self.PrintProcess(number)
    self._printer.Println()

  def PrintProcess(self, number):
    process = self.GetProcessFromNumber(number)
    if process==None: return
    self._printer.PrintProcess(process)

  def PrintNextLocationsOfProcesses(self, numbers):
    for number in numbers:
      self.PrintNextLocationsOfProcess(number)
    self._printer.Println()

  def PrintNextLocationsOfProcess(self, number):
    process = self.GetProcessFromNumber(number)
    if process==None: return
    self._printer.PrintNextLocationsOfProcess(process)

  def PrintHistory(self):
    history = self.GetHistory()
    self._printer.PrintHistory(history)

# 10) Messages
# ------------

  def GetOutputFile(self):
    return self._outputFile
  def SetOutputFile(self, file):
    self._outputFile = file
    self._printer.SetOutputFile(file)

  def GetErrorFile(self):
    return self._errorFile
  def SetErrorFile(self, file):
    self._errorFile = file
    self._parser.SetErrorFile(file)
    self._rewriter.SetErrorFile(file)

  def IsVerbose(self):
    return self._verbose
  def SetVerbose(self, value):
    self._verbose = value

  def Print(self, *texts):
    self._printer.Print(*texts)
  def Println(self, *texts):
    self._printer.Println(*texts)

  def PrintMessage(self, message):
    if not self._verbose:
      return
    self.Println(message)

  def RaiseError(self, message):
    if not self._verbose:
      return
    self._errorFile.write(message + "\n")
    self._errorFile.flush()
