# Scola version 1.1.0
# ScolaModels.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Models
#  2.1) Models
#  2.2) Classes
#  2.3) Instances
#  2.4) Clones Directives
#  2.5) Inherits Directives
# 3) System
#  3.1) Domains
#  3.2) Blocks
#  3.3) Ports
#  3.4) Assertions
# 4) Use cases
#  4.1) Scenarios
#  4.2) States
#  4.3) Tasks
#  4.4) Test gateways
#   4.4.1) Gateway cases
#  4.5) Choice gateways
#   4.5.1) Gateway branches
#  4.6) Fork gateways
#  4.7) Join gateways
#  4.8) Split gateways
#  4.9) Merge gateways
#  4.10) Meet gateways
#  4.11) Next directives
# 5) Instructions
#  5.1) Generic instructions
#  5.2) fail instruction
#  5.3) set instruction
#  5.4) new_port instruction
#  5.5) new_block instruction
#  5.6) delete instruction
#  5.7) move instruction
#  5.8) send instruction
#  5.9) receive instruction
#  5.10) clone instruction
#  5.11) if instruction
#  5.12) Instruction block
#  5.13) print instruction

# 1) Imported modules
# -------------------

import string
import sys

from ScolaMetaModel import *
from ScolaUniverse import *
from ScolaExpressions import *

# 2) Models
# ---------

# 2.1) Models
# -----------

class ScolaModel(ScolaContainer):
  UNKNOWN = "UNKNOWN"
  MODEL = "MODEL"
  CLASS = "CLASS"
  INSTANCE = "INSTANCE"
  CLONES = "CLONES"
  INHERITS = "INHERITS"
  DOMAIN = "DOMAIN"
  BLOCK = "BLOCK"
  PORT = "PORT"
  ASSERTION = "ASSERTION"
  SCENARIO = "SCENARIO"
  STATE = "STATE"
  TASK = "TASK"
  TEST = "TEST"
  CASE = "CASE"
  CHOICE = "CHOICE"
  BRANCH = "BRANCH"
  FORK = "FORK"
  JOIN = "JOIN"
  SPLIT = "SPLIT"
  MERGE = "MERGE"
  MEET = "MEET"
  NEXT = "NEXT"
  FAIL = "FAIL"
  SET = "SET"
  NEW_PORT = "NEW_PORT"
  NEW_BLOCK = "NEW_BLOCK"
  DELETE = "DELETE"
  MOVE = "MOVE"
  SEND = "SEND"
  RECEIVE = "RECEIVE"
  CLONE = "CLONE"
  IF = "IF"
  BEGIN = "BEGIN"
  PRINT = "PRINT"

  def __init__(self):
    ScolaContainer.__init__(self, ScolaModel.MODEL, None)
    self._isChecked = False

  def IsChecked(self):
    return self._isChecked
  def SetChecked(self, value):
    self._isChecked = value

  @staticmethod
  def IsScenarioElement(element):
    if element.GetType()!=ScolaModel.STATE: return True
    if element.GetType()!=ScolaModel.TASK: return True
    if element.GetType()!=ScolaModel.TEST: return True
    if element.GetType()!=ScolaModel.CASE: return True
    if element.GetType()!=ScolaModel.CHOICE: return True
    if element.GetType()!=ScolaModel.BRANCH: return True
    if element.GetType()!=ScolaModel.FORK: return True
    if element.GetType()!=ScolaModel.JOIN: return True
    if element.GetType()!=ScolaModel.SPLIT: return True
    if element.GetType()!=ScolaModel.MERGE: return True
    return False

# 2.2) Classes
# ------------

class ScolaClass(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.CLASS, container)
    self.Rename(identifier)
    self._eventualType = ScolaModel.UNKNOWN
    self._identifierOfInheritedClass = None
    self._inheritedClass = None

  def GetEventualType(self):
    return self._eventualType
  def SetEventualType(self, type):
    self._eventualType = type

  def GetIdentifierOfInheritedClass(self):
    return self._identifierOfInheritedClass
  def SetIdentifierOfInheritedClass(self, identifier):
    self._identifierOfInheritedClass = identifier 

  def GetInheritedClass(self):
    return self._inheritedClass
  def SetInheritedClass(self, inheritedClass):
    self._inheritedClass = inheritedClass 

# 2.3) Instances
# --------------

class ScolaInstance(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.INSTANCE, container)
    self.Rename(identifier)
    self._eventualType = ScolaModel.UNKNOWN
    self._classIdentifier = None
    self._classDeclaration = None
    self._asPath = None

  def GetEventualType(self):
    return self._eventualType
  def SetEventualType(self, type):
    self._eventualType = type

  def GetClassIdentifier(self):
    return self._classIdentifier
  def SetClassIdentifier(self, classIdentifier):
    self._classIdentifier = classIdentifier

  def GetClassDeclaration(self):
    return self._classDeclaration
  def SetClassDeclaration(self, classDeclaration):
    self._classDeclaration = classDeclaration

  def GetAsPath(self):
    return self._asPath
  def SetAsPath(self, path):
    self._asPath = path

# 2.4) Clones Directive
# ---------------------

class ScolaClonesDirective(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.CLONES, container)
    self.Rename(identifier)
    self._eventualType = ScolaModel.UNKNOWN
    self._pathToClonedElement = None
    self._clonedElement = None
    self._asPath = None

  def GetEventualType(self):
    return self._eventualType
  def SetEventualType(self, type):
    self._eventualType = type

  def GetPathToClonedElement(self):
    return self._pathToClonedElement
  def SetPathToClonedElement(self, pathToClonedElement):
    self._pathToClonedElement = pathToClonedElement 

  def GetClonedElement(self):
    return self._clonedElement
  def SetClonedElement(self, clonedElement):
    self._clonedElement = clonedElement 

  def GetAsPath(self):
    return self._asPath
  def SetAsPath(self, path):
    self._asPath = path

# 2.5) Inherits Directive
# -----------------------

class ScolaInheritsDirective(ScolaElement):
  def __init__(self, container):
    ScolaElement.__init__(self, ScolaModel.INHERITS, container)
    self._eventualType = ScolaModel.UNKNOWN
    self._pathToInheritedElement = None
    self._inheritedElement = None

  def GetEventualType(self):
    return self._eventualType
  def SetEventualType(self, type):
    self._eventualType = type

  def GetPathToInheritedElement(self):
    return self._pathToInheritedElement
  def SetPathToInheritedElement(self, pathToInheritedElement):
    self._pathToInheritedElement = pathToInheritedElement 

  def GetInheritedElement(self):
    return self._inheritedElement
  def SetInheritedElement(self, inheritedElement):
    self._inheritedElement = inheritedElement 

# 3) System
# ---------

# 3.1) Domains
# ------------

class ScolaDomain(ScolaElement):
  def __init__(self, identifier, container):
    ScolaElement.__init__(self, ScolaModel.DOMAIN, container)
    self.Rename(identifier)
    self._domain = None

  def GetDomain(self):
    return self._domain
  def SetDomain(self, domain):
    self._domain = domain

# 3.2) Blocks
# -----------

class ScolaBlock(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.BLOCK, container)
    self.Rename(identifier)
    self._receptionSet = dict()

  def CloneBlock(self, sourceBlock, identifier):
    targetBlock = ScolaBlock(identifier, self)
    for element in sourceBlock.GetElements():
      if element.GetType()==ScolaModel.BLOCK:
        targetBlock.CloneBlock(element, element.GetIdentifier())
      elif element.GetType()==ScolaModel.PORT:
        targetBlock.ClonePort(element, element.GetIdentifier())
    return targetBlock

  def ClonePort(self, sourcePort, identifier):
    sourceInitialization = sourcePort.GetInitialization()
    targetInitialization = sourceInitialization.DeepCopy()
    sourceValue = sourcePort.GetValue()
    targetValue = ScolaConstant.Copy(sourceValue)
    targetPort = ScolaPort(identifier, targetInitialization, self)
    targetPort.SetValue(targetValue)
    return targetPort

  def IsReceptionSetEmpty(self):
    return len(self._receptionSet)==0
  def ResetReceptionSet(self):
    self._receptionSet = dict()
  def LookForElementInReceptionSet(self, key):
    return self._receptionSet.get(key, None)
  def AddElementToReceptionSet(self, key, element):
    self._receptionSet[key] = element
  def RemoveElementFromReceptionSet(self, key):
    del self._receptionSet[key]

# 3.3) Ports
# ----------

class ScolaPort(ScolaElement):
  def __init__(self, baseType, identifier, expression, container):
    ScolaElement.__init__(self, ScolaModel.PORT, container)
    self._baseType = baseType
    self.Rename(identifier)
    self._initialization = expression
    self._value = ScolaConstant.NewNone()

  def GetBaseType(self):
    return self._baseType
  def SetBaseType(self, baseType):
    self._baseType = baseType

  def GetInitialization(self):
    return self._initialization
  def SetInitialization(self, expression):
    self._initialization = expression

  def GetValue(self):
    return self._value
  def SetValue(self, value):
    self._value = value


# 3.4) Assertions
# ---------------

class ScolaAssertion(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.ASSERTION, container)
    self.Rename(identifier)

# 4) Use cases
# ------------

# 4.1) Scenarios
# --------------

class ScolaScenario(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.SCENARIO, container)
    self.Rename(identifier)
    self._asPath = None

  def GetAsPath(self):
    return self._asPath
  def SetAsPath(self, path):
    self._asPath = path

# 4.2) States
# -----------

class ScolaState(ScolaElement):
  def __init__(self, identifier, container):
    ScolaElement.__init__(self, ScolaModel.STATE, container)
    self.Rename(identifier)
    self._isInitial = True

  def IsInitial(self):
    return self._isInitial
  def SetInitial(self, value):
    self._isInitial = value

# 4.3) Tasks
# ----------

class ScolaTask(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.TASK, container)
    self.Rename(identifier)

# 4.4) Test gateways
# ------------------

class ScolaTestGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.TEST, container)
    self.Rename(identifier)

# 4.4.1) Gateway cases
# --------------------

class ScolaGatewayCase(ScolaElement):
  def __init__(self, identifier, condition, container):
    ScolaElement.__init__(self, ScolaModel.CASE, container)
    self.Rename(identifier)
    self._condition = condition

  def GetCondition(self):
    return self._condition

# 4.5) Choice gateways
# --------------------

class ScolaChoiceGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.CHOICE, container)
    self.Rename(identifier)

# 4.5.1) Gateway branches
# -----------------------

class ScolaGatewayBranch(ScolaElement):
  def __init__(self, identifier, container):
    ScolaElement.__init__(self, ScolaModel.BRANCH, container)
    self.Rename(identifier)
    self._queue = []

  def IsQueueEmpty(self):
    return len(self._queue)==0

  def GetIndexInQueue(self, process):
    return self._queue.index(process)

  def FirstInQueue(self):
    return self._queue[0]

  def EnqueueProcess(self, process):
    self._queue.append(process)

  def DequeueProcess(self):
    return self._queue.pop(0)

  def RemoveProcess(self, process):
    index = self._queue.index(process)
    self._queue.remove(process)
    return index

  def InsertProcess(self, index, process):
    self._queue.insert(index, process)

  def GetQueue(self):
    return self._queue

# 4.6) Fork gateways
# ------------------

class ScolaForkGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.FORK, container)
    self.Rename(identifier)

# 4.7) Join gateways
# ------------------

class ScolaJoinGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.JOIN, container)
    self.Rename(identifier)

# 4.8) Split gateways
# ------------------

class ScolaSplitGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.SPLIT, container)
    self.Rename(identifier)

# 4.9) Merge gateways
# -------------------

class ScolaMergeGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.MERGE, container)
    self.Rename(identifier)
    self._queues = dict()

# 4.10) Meet gateways
# ------------------

class ScolaMeetGateway(ScolaContainer):
  def __init__(self, identifier, container):
    ScolaContainer.__init__(self, ScolaModel.MEET, container)
    self.Rename(identifier)

# 4.11) Next directives
# ---------------------

class ScolaNextDirective(ScolaElement):
  def __init__(self, sourcePath, targetPath, container):
    ScolaElement.__init__(self, ScolaModel.NEXT, container)
    self.SetPath(sourcePath)
    self._targetPath = targetPath

  def GetSourcePath(self):
    return self.GetPath()
  def GetTargetPath(self):
    return self._targetPath


# 5) Instructions
# ---------------

# 5.1) Generic instructions
# -------------------------

class ScolaInstruction(ScolaElement):
  def __init__(self, type, pathExpression, container):
    ScolaElement.__init__(self, type, container)
    self._pathExpression = pathExpression

  def GetPathExpression(self):
    return self._pathExpression
  def SetPathExpression(self, pathExpression):
    self._pathExpression = pathExpression

# 5.2) fail instruction
# --------------------

class ScolaFailInstruction(ScolaInstruction):
  def __init__(self, container):
    ScolaInstruction.__init__(self, ScolaModel.FAIL, None, container)


# 5.3) set instruction
# --------------------

class ScolaSetInstruction(ScolaInstruction):
  def __init__(self, pathExpression, valueExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.SET, pathExpression, container)
    self._valueExpression = valueExpression

  def GetValueExpression(self):
    return self._valueExpression

# 5.4) new_port instruction
# -------------------------

class ScolaNewPortInstruction(ScolaInstruction):
  def __init__(self, baseType, pathExpression, initializationExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.NEW_PORT, pathExpression, container)
    self._baseType = baseType
    self._initializationExpression = initializationExpression

  def GetBaseType(self):
    return self._baseType

  def GetInitializationExpression(self):
    return self._initializationExpression

# 5.5) new_block instruction
# --------------------------

class ScolaNewBlockInstruction(ScolaInstruction):
  def __init__(self, blockClass, pathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.NEW_BLOCK, pathExpression, container)
    self._blockClass = blockClass

  def GetBlockClass(self):
    return self._blockClass

# 5.6) delete instruction
# -----------------------

class ScolaDeleteInstruction(ScolaInstruction):
  def __init__(self, pathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.DELETE, pathExpression, container)

# 5.7) move instruction
# ---------------------

class ScolaMoveInstruction(ScolaInstruction):
  def __init__(self, pathExpression, targetPathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.MOVE, pathExpression, container)
    self._targetPathExpression = targetPathExpression

  def GetTargetPathExpression(self):
    return self._targetPathExpression

# 5.8) send instruction
# ---------------------

class ScolaSendInstruction(ScolaInstruction):
  def __init__(self, pathExpression, receiverPathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.SEND, pathExpression, container)
    self._receiverPathExpression = receiverPathExpression

  def GetReceiverPathExpression(self):
    return self._receiverPathExpression

# 5.9) receive instruction
# ------------------------

class ScolaReceiveInstruction(ScolaInstruction):
  def __init__(self, pathExpression, senderPathExpression, targetPathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.RECEIVE, pathExpression, container)
    self._senderPathExpression = senderPathExpression
    self._targetPathExpression = targetPathExpression

  def GetSenderPathExpression(self):
    return self._senderPathExpression
  def GetTargetPathExpression(self):
    return self._targetPathExpression

# 5.10) clone instruction
# -----------------------

class ScolaCloneInstruction(ScolaInstruction):
  def __init__(self, pathExpression, targetPathExpression, container):
    ScolaInstruction.__init__(self, ScolaModel.CLONE, pathExpression, container)
    self._targetPathExpression = targetPathExpression

  def GetTargetPathExpression(self):
    return self._targetPathExpression


# 5.11) if instruction
# --------------------

class ScolaIfInstruction(ScolaContainer):
  def __init__(self, condition, container):
    ScolaContainer.__init__(self, ScolaModel.IF, container)
    self._condition = condition

  def GetCondition(self):
    return self._condition
  def GetThenInstruction(self):
    return self.GetElement(0)
  def GetElseInstruction(self):
    return self.GetElement(1)

# 5.12) Instruction block
# -----------------------

class ScolaInstructionBlock(ScolaContainer):
  def __init__(self, container):
    ScolaContainer.__init__(self, ScolaModel.BEGIN, container)

# 5.13) print instruction
# -----------------------

class ScolaPrintInstruction(ScolaInstruction):
  def __init__(self, expression, container):
    ScolaInstruction.__init__(self, ScolaModel.PRINT, expression, container)

